import os.path
import sys

import pytest

from bonsai import LDAPClient
from bonsai import LDAPEntry
from bonsai import LDAPModOp
from bonsai import LDAPDN
import bonsai.errors
from bonsai.errors import AuthenticationError, InvalidDN
from bonsai.active_directory import UserAccountControl


@pytest.fixture
def test_entry():
    """Create a test LDAP entry."""
    gconn = None
    entry = None

    def _create_entry(conn, dname):
        nonlocal gconn
        nonlocal entry
        entry = LDAPEntry(dname)
        entry["objectclass"] = [
            "top",
            "inetOrgPerson",
            "person",
            "organizationalPerson",
        ]
        entry["sn"] = "test"
        gconn = conn
        try:
            conn.add(entry)
        except bonsai.AlreadyExists:
            conn.delete(entry.dn)
            conn.add(entry)
        return entry

    yield _create_entry

    if gconn.closed:
        gconn = gconn.open()
    gconn.delete(entry.dn)
    gconn.close()


@pytest.fixture
def test_ad_user():
    """Create a test AD user."""
    gconn = None
    entry = None

    def _create_entry(conn, dname, password):
        nonlocal gconn
        nonlocal entry
        entry = LDAPEntry(dname)
        entry["objectclass"] = [
            "top",
            "inetOrgPerson",
            "person",
            "organizationalPerson",
            "user",
        ]
        entry["sn"] = "ad_user"
        entry["cn"] = "ad_user"
        entry["givenName"] = "AD"
        entry["userPrincipalName"] = "ad_user"
        entry["displayName"] = "ad_user"
        entry["sAMAccountName"] = "ad_user"
        entry["userPassword"] = password
        entry["mail"] = "ad_user@bonsai.test"
        gconn = conn
        try:
            conn.add(entry)
        except bonsai.AlreadyExists:
            conn.delete(entry.dn)
            conn.add(entry)
        return entry

    yield _create_entry

    if gconn.closed:
        gconn = gconn.open()
    gconn.delete(entry.dn)
    gconn.close()


def test_set_get():
    """Test LDAPEntry's SetItem, GetItem and get methods."""
    entry = LDAPEntry("cn=test")
    entry["sn"] = "Test"
    assert entry == {"dn": LDAPDN("cn=test"), "sn": ["Test"]}
    entry["givenname"] = "Test"
    assert entry.get("None") is None
    assert entry.get("GivenName") == entry["givenNAME"]
    del entry["sn"]
    with pytest.raises(KeyError):
        _ = entry["sn"]


def test_append_extend():
    """Test append and extend methods of LDAPEntry's attribute."""
    entry = LDAPEntry("cn=test")
    entry["givenName"] = "test"
    entry["givenname"].append("test2")
    assert entry["givenname"] == ["test", "test2"]
    assert entry["givenname"][0] == "test"
    with pytest.raises(ValueError):
        entry["GivenName"].extend(["teSt", "test3"])


def test_pop():
    """Test LDAPEntry's pop method."""
    entry = LDAPEntry("cn=test")
    entry["test"] = "test"
    with pytest.raises(TypeError):
        _ = entry.pop()
    with pytest.raises(TypeError):
        _ = entry.pop("t", 2, 3)
    with pytest.raises(KeyError):
        _ = entry.pop("t")
    assert entry.pop("test") == ["test"]
    assert entry.pop("test", None) is None


def test_popitem():
    """Test LDAPEntry's popitem method."""
    entry = LDAPEntry("cn=test")
    entry["test"] = "test"
    entry["test2"] = "test"
    item = entry.popitem()
    assert len(item) == 2
    assert item[0] not in entry
    entry[item[0]] = item[1]
    assert entry[item[0]] == item[1]


def test_popitem_empty():
    """Test LDAPEntry's popitem raises KeyError if it is empty."""
    entry = LDAPEntry("cn=test")
    with pytest.raises(KeyError):
        _ = entry.popitem()


def test_clear():
    """Test LDAPEntry's clear method."""
    entry = LDAPEntry("cn=test")
    entry["sn"] = ["test1", "test2"]
    entry["gn"] = ["test3"]
    entry.clear()
    assert entry == {"dn": LDAPDN("cn=test")}
    assert entry.dn == "cn=test"


def test_items():
    """Test LDAPEntry's items method."""
    entry = LDAPEntry("cn=test")
    entry["cn"] = "test"
    entry["sn"] = "Test"
    assert len(entry.items()) == 3
    assert ("dn", entry.dn) in entry.items()
    assert ("cn", entry["cn"]) in entry.items()
    assert ("sn", entry["sn"]) in entry.items()
    assert len(list(entry.items(exclude_dn=True))) == 2
    assert ("dn", entry.dn) not in entry.items(exclude_dn=True)
    assert ("cn", entry["cn"]) in entry.items(exclude_dn=True)
    assert ("sn", entry["sn"]) in entry.items(exclude_dn=True)


def test_keys():
    """Test LDAPEntry's keys method."""
    entry = LDAPEntry("cn=test")
    entry["cn"] = "test"
    entry["sn"] = "Test"
    assert set(entry.keys()) == set(["dn", "cn", "sn"])
    assert set(entry.keys(exclude_dn=True)) == set(["cn", "sn"])


def test_values():
    """Test LDAPEntry's values method."""
    entry = LDAPEntry("cn=test")
    entry["cn"] = "test"
    entry["sn"] = "Test"
    assert len(entry.values()) == 3
    assert entry.dn in entry.values()
    assert entry["cn"] in entry.values()
    assert entry["sn"] in entry.values()
    assert len(list(entry.values(exclude_dn=True))) == 2
    assert entry.dn not in entry.values(exclude_dn=True)
    assert entry["cn"] in entry.values(exclude_dn=True)
    assert entry["sn"] in entry.values(exclude_dn=True)


def test_update():
    """Test updating LDAPEntry object."""
    entry = LDAPEntry("cn=test")
    entry.update({"GivenName": "test2", "mail": "test@mail"})
    entry.update([("sn", "test")])
    with pytest.raises(ValueError):
        entry.update([("sn", "test", 1)])
    entry.update(uidnumber=1, gidnumber=1)
    assert entry["mail"] == ["test@mail"]
    assert entry["givenname"] == ["test2"]
    assert entry["sn"][0] == "test"
    assert entry["uidnumber"] == [1]
    assert entry["gidnumber"] == [1]


def test_equal():
    """Test equality check."""
    entry1 = LDAPEntry("cn=test")
    entry2 = LDAPEntry("cn=test")
    entry3 = LDAPEntry("cn=test1")
    assert entry1 == entry2
    assert not (entry1 == entry3)
    assert entry1 == {"dn": LDAPDN("cn=test")}
    assert not (entry1 == 2)


def test_special_char(client, basedn):
    """Test adding entry with special character in its DN."""
    with client.connect() as conn:
        entry = LDAPEntry(r"cn=test\, *\+withspec,%s" % basedn)
        entry["objectclass"] = ["top", "inetOrgPerson"]
        entry["sn"] = "Test,*special"
        conn.add(entry)
        result = conn.search(basedn, 1)
        entry.delete()
        assert entry.dn in [res.dn for res in result]


def test_dn_with_space(client, basedn):
    """Test adding entry with DN that has a space in it."""
    with client.connect() as conn:
        entry = LDAPEntry("cn=test, %s" % basedn)
        entry["objectclass"] = ["top", "inetOrgPerson"]
        entry["sn"] = "Test with space"
        conn.add(entry)
        result = conn.search(basedn, 1)
        entry.delete()
        assert " " in str(entry.dn)
        assert entry.dn.rdns[0][0][1] in [res.dn.rdns[0][0][1] for res in result]


def test_unicode(client, basedn):
    """Test adding entry with special character in its DN."""
    with client.connect() as conn:
        dname = "cn=test_µčབྷñ,%s" % basedn
        entry = LDAPEntry(dname)
        entry["objectclass"] = ["top", "inetOrgPerson"]
        entry["sn"] = "unicode_µčབྷñ"
        conn.add(entry)
        result = conn.search(dname, 0)
        entry.delete()
        assert dname in [res.dn for res in result]


def test_binary(client, basedn):
    """Test adding binary data."""
    curdir = os.path.abspath(os.path.dirname(__file__))
    conn = client.connect()
    dname = "cn=binary,%s" % basedn
    entry = LDAPEntry(dname)
    entry["objectclass"] = ["top", "inetOrgPerson"]
    entry["sn"] = "binary_test"
    with open("%s/testenv/test.jpeg" % curdir, "rb") as image:
        entry["jpegPhoto"] = image.read()
    conn.add(entry)
    result = conn.search(dname, 0)
    entry.delete()
    conn.close()
    assert "jpegPhoto" in result[0].keys()
    assert result[0]["jpegphoto"][0] == entry["jpegphoto"][0]


def test_connection(client, basedn):
    """Test set and get connection object from LDAPEntry."""
    entry = LDAPEntry("cn=test,%s" % basedn)
    with pytest.raises(ValueError):
        _ = entry.connection
    conn = client.connect()
    entry.connection = conn
    assert entry.connection == conn
    with pytest.raises(TypeError):
        entry.connection = "string"
    with pytest.raises(TypeError):
        del entry.connection


def test_rename(client, basedn, test_entry):
    """Test LDAPEntry's rename LDAP operation."""
    with client.connect() as conn:
        entry = test_entry(conn, "cn=test,%s" % basedn)
        entry.rename("cn=test2,%s" % basedn)
        assert str(entry.dn) == "cn=test2,%s" % basedn
        obj = conn.search("cn=test,%s" % basedn, 0)
        assert obj == []
        obj = conn.search("cn=test2,%s" % basedn, 0)[0]
        assert entry.dn == obj.dn


@pytest.mark.skipif(
    sys.platform.startswith("win"), reason="Cannot rename entry with old RDN on Windows"
)
def test_rename_with_old_rdn(client, basedn, test_entry):
    """Test LDAPEntry's rename LDAP operation."""
    with client.connect() as conn:
        entry = test_entry(conn, "cn=test,%s" % basedn)
        entry.rename("uid=test2,%s" % basedn, delete_old_rdn=False)
        assert str(entry.dn) == "uid=test2,%s" % basedn
        obj = conn.search("cn=test,%s" % basedn, 0)
        assert obj == []
        obj = conn.search("uid=test2,%s" % basedn, 0)[0]
        assert entry.dn == obj.dn
        assert "test2" in obj["uid"]
        assert "test" in obj["cn"]


def test_rename_error(client, basedn, test_entry):
    """Test LDAPEntry's rename error handling."""
    dname = bonsai.LDAPDN("cn=test,%s" % basedn)
    entry = LDAPEntry(dname)
    with pytest.raises(ValueError):
        entry.rename("cn=test2")
    with client.connect() as conn:
        entry = test_entry(conn, dname)
        with pytest.raises(TypeError):
            entry.rename(0)
        with pytest.raises(TypeError):
            entry.rename(0, delete_old_rdn=0)
        try:
            newdn = bonsai.LDAPDN("cn=test2,ou=invalid,%s" % basedn)
            entry.rename(newdn)
        except bonsai.LDAPError:
            assert entry.dn == dname
    with pytest.raises(bonsai.errors.ClosedConnection):
        entry.rename("cn=test2")


def test_sync_operations(client, basedn):
    """
    Test LDAPEntry's add, modify and delete synchronous operations.
    """
    entry = LDAPEntry("cn=test,%s" % basedn)
    with pytest.raises(ValueError):
        entry.modify()
    with client.connect() as conn:
        entry["sn"] = "test"
        with pytest.raises(bonsai.ObjectClassViolation):
            conn.add(entry)
        entry["objectclass"] = [
            "top",
            "inetOrgPerson",
            "person",
            "organizationalPerson",
        ]
        try:
            conn.add(entry)
        except bonsai.AlreadyExists:
            conn.delete(entry.dn)
            conn.add(entry)
        except:
            pytest.fail("Adding LDAPEntry to the server is failed.")
        entry["sn"] = "Test_modify"
        try:
            entry.modify()
        except:
            pytest.fail("Modify failed.")
        obj = conn.search("cn=test,%s" % basedn, 0)[0]
        assert entry["sn"] == obj["sn"]
        try:
            entry.delete()
        except:
            pytest.fail("Delete failed.")
    with pytest.raises(bonsai.errors.ClosedConnection):
        entry.modify()


def test_dn_attr(basedn):
    """Test LDAPEntry's DN attribute."""
    entry = LDAPEntry("cn=test,%s" % basedn)
    entry.dn = "cn=test"
    assert str(entry.dn) == "cn=test"
    with pytest.raises(TypeError):
        del entry.dn
    with pytest.raises(TypeError):
        entry["dn"] = 5


def test_wrong_params():
    """Test passing wrong params to LDAPEntry."""
    with pytest.raises(TypeError):
        _ = LDAPEntry("", 1)
    with pytest.raises(InvalidDN):
        _ = LDAPEntry("5", 1)


@pytest.mark.skipif(
    sys.platform.startswith("win"), reason="Cannot use password policy on Windows"
)
def test_password_modify(client):
    """
    Test modifying password with simple modify operation and
    password policy.
    """
    cli = LDAPClient(client.url)
    user_dn = "cn=jeff,ou=nerdherd,dc=bonsai,dc=test"
    cli.set_password_policy(True)
    cli.set_credentials("SIMPLE", user_dn, "p@ssword")
    conn, _ = cli.connect()
    entry = conn.search(user_dn, 0)[0]
    try:
        entry["userPassword"] = "newpassword"
        entry.modify()
    except Exception as exc:
        assert isinstance(exc, bonsai.errors.PasswordModNotAllowed)
    user_dn = "cn=skip,ou=nerdherd,dc=bonsai,dc=test"
    cli.set_credentials("SIMPLE", user_dn, "p@ssword")
    conn, _ = cli.connect()
    entry = conn.search(user_dn, 0)[0]
    try:
        entry["userPassword"] = "short"
        entry.modify()
    except Exception as exc:
        assert isinstance(exc, bonsai.errors.PasswordTooShort)
    try:
        entry["userPassword"] = "p@ssword"
        entry.modify()
    except Exception as exc:
        assert isinstance(exc, bonsai.errors.PasswordInHistory)


def test_extended_dn_attr():
    """Test that extended dn attribute is read only."""
    entry = LDAPEntry("cn=test")
    with pytest.raises(AttributeError):
        entry.extended_dn = "cn=test2"


def test_change_attribute(client):
    """Test change_attribute method."""
    user_dn = "cn=sam,ou=nerdherd,dc=bonsai,dc=test"
    if sys.platform == "win32":
        multiattr = "otherTelephone"
    else:
        multiattr = "mail"
    with client.connect() as conn:
        entry = LDAPEntry(user_dn, conn)
        entry.change_attribute("mail", LDAPModOp.ADD, "sam@bonsai.test")
        assert entry["mail"].status == 1
        entry.modify()
        assert conn.search(user_dn, 0)[0]["mail"][0] == "sam@bonsai.test"
        entry.change_attribute("mail", 1, "sam@bonsai.test")
        assert entry["mail"].status == 1
        entry.modify()
        with pytest.raises(KeyError):
            _ = conn.search(user_dn, 0)[0]["mail"]
        entry.change_attribute(
            multiattr, LDAPModOp.REPLACE, "sam@bonsai.test", "x@bonsai.test"
        )
        assert entry[multiattr].status == 2
        entry.modify()
        res = conn.search(user_dn, 0)[0][multiattr]
        assert "sam@bonsai.test" in res
        assert "x@bonsai.test" in res
        entry.change_attribute(multiattr, 1, "x@bonsai.test")
        entry.change_attribute(multiattr, 0, "sam2@bonsai.test")
        entry.modify()
        res = conn.search(user_dn, 0)[0][multiattr]
        assert "sam@bonsai.test" in res
        assert "sam2@bonsai.test" in res
        entry.change_attribute(multiattr, 1)
        entry.modify()
        assert multiattr not in conn.search(user_dn, 0)[0].keys()


def test_change_attribute_error(client):
    """Test change_attribute method's error handling."""
    user_dn = "cn=sam,ou=nerdherd,dc=bonsai,dc=test"
    with client.connect() as conn:
        entry = LDAPEntry(user_dn, conn)
        with pytest.raises(ValueError):
            entry.change_attribute("mail", 4, "t")
        entry.change_attribute("sn", 0, "Lembeck")
        with pytest.raises(bonsai.TypeOrValueExists):
            entry.modify()
        entry.clear_attribute_changes("sn")
        entry.change_attribute("description", LDAPModOp.DELETE, "sam")
        with pytest.raises(bonsai.NoSuchAttribute):
            entry.modify()


def test_clear_attribute_changes():
    """Test clear_attribute_changes method."""
    user_dn = "cn=sam,ou=nerdherd,dc=bonsai,dc=test"
    entry = LDAPEntry(user_dn)
    entry.change_attribute("uidNumber", 0, 4)
    assert entry["uidNumber"].added == [4]
    entry.change_attribute("uidNumber", 1, 4)
    assert entry["uidNumber"].deleted == [4]
    entry.clear_attribute_changes("uidNumber")
    assert entry["uidNumber"].status == 0
    assert entry["uidNumber"].added == []
    assert entry["uidNumber"].deleted == []


@pytest.mark.skipif(
    sys.platform.startswith("win"), reason="Cannot use ManageDsaIT on Windows"
)
def test_modify_referrals(client):
    """Test modifying an LDAP referral with ManageDdsIT control."""
    refdn = bonsai.LDAPDN("o=invalid-ref,ou=nerdherd-refs,dc=bonsai,dc=test")
    newref = "ldap://invalid.host/cn=nobody"
    cli = LDAPClient(client.url)
    cli.set_credentials(client.mechanism, **client.credentials)
    cli.managedsait = True
    with cli.connect() as conn:
        entry = LDAPEntry(refdn, conn)
        entry.change_attribute("ref", LDAPModOp.ADD, newref)
        entry.modify()
        res = conn.search(refdn, 0, attrlist=["ref"])[0]
        assert len(res["ref"]) == 3
        assert newref in res["ref"]
        entry.change_attribute("ref", LDAPModOp.DELETE, newref)
        entry.modify()


@pytest.mark.skipif(
    not sys.platform.startswith("win"),
    reason="Makes sense when the remote server is an Active Directory",
)
def test_create_ad_user(cfg, basedn, test_ad_user):
    """Test creating and enabling an AD user ."""
    url = "ldap://%s:%s/%s??%s" % (
        cfg["SERVER"]["hostname"],
        cfg["SERVER"]["port"],
        cfg["SERVER"]["basedn"],
        cfg["SERVER"]["search_scope"],
    )
    cli = LDAPClient(url, tls=True)
    cli.set_ca_cert("./tests/testenv/certs/cacert.pem")
    cli.set_credentials(
        "SIMPLE", user=cfg["SIMPLEAUTH"]["user"], password=cfg["SIMPLEAUTH"]["password"]
    )
    ad_user_dn = "cn=ad_user,%s" % basedn
    password = "C0mpleX_P4ssW0rd"
    with cli.connect() as conn:
        entry = test_ad_user(conn, ad_user_dn, password)
        res = conn.search(entry.dn, 0)
        assert res
        uac = UserAccountControl(res[0]["userAccountControl"][0])
        assert uac.value != 66048
        assert uac.properties["accountdisable"]
        user_cli = LDAPClient(url, tls=True)
        user_cli.set_credentials("SIMPLE", ad_user_dn, password)
        with pytest.raises(AuthenticationError):
            _ = user_cli.connect()
        entry.change_attribute("userAccountControl", LDAPModOp.REPLACE, 66048)
        entry.modify()
        with user_cli.connect() as uconn:
            res = uconn.search(entry.dn, 0)
            assert not UserAccountControl(res[0]["userAccountControl"][0]).properties[
                "accountdisable"
            ]
            assert uconn.whoami() == "u:BONSAI\\ad_user"



import logging
import threading
from contextlib import contextmanager
from typing import Optional, Any, Set, Generic, TypeVar, Generator

from .ldapconnection import BaseLDAPConnection, LDAPConnection

MYPY = False

if MYPY:
    from .ldapclient import LDAPClient


logger = logging.getLogger("bonsai.pool")


class PoolError(Exception):
    """Connection pool related errors."""

    pass


class ClosedPool(PoolError):
    """Raised, when the connection pool is closed."""

    pass


class EmptyPool(PoolError):
    """Raised, when the connection pool is empty."""

    pass


T = TypeVar("T", bound=BaseLDAPConnection)


class ConnectionPool(Generic[T]):
    """
    A connection pool object for managing multiple open connections.

    :param LDAPClient client: the :class:`bonsai.LDAPClient` that's used to create
                connections.
    :param int minconn: the minimum number of connections that's created
                after the pool is opened.
    :param int maxconn: the maximum number of connections in the pool.
    :param \\*\\*kwargs: additional keyword arguments that are passed to
                the :meth:`bonsai.LDAPClient.connect` method.
    :raises ValueError: when the minconn is negative or the maxconn is less
        than the minconn.
    """

    def __init__(
        self, client: "LDAPClient", minconn: int = 1, maxconn: int = 10, **kwargs: Any
    ) -> None:
        """Init method."""
        if minconn < 0:
            raise ValueError("The minconn must be positive.")
        if minconn > maxconn:
            raise ValueError("The maxconn must be greater than minconn.")
        self._minconn = minconn
        self._maxconn = maxconn
        self._client = client
        self._kwargs = kwargs
        self._closed = True
        self._idles: Set[T] = set()
        self._used: Set[T] = set()

    def open(self) -> None:
        """
        Open the connection pool by initialising the minimal number of
        connections.
        """
        for _ in range(self._minconn - self.idle_connection - self.shared_connection):
            self._idles.add(self._client.connect(**self._kwargs))
        self._closed = False

    def get(self) -> T:
        """
        Get a connection from the connection pool.

        :raises EmptyPool: when the pool is empty.
        :raises ClosedPool: when the method is called on a closed pool.
        :return: an LDAP connection object.
        """
        if self._closed:
            raise ClosedPool("The pool is closed.")
        try:
            conn = self._idles.pop()
        except KeyError:
            if len(self._used) < self._maxconn:
                conn = self._client.connect(**self._kwargs)
            else:
                raise EmptyPool("Pool is empty.") from None
        self._used.add(conn)
        return conn

    def put(self, conn: T) -> None:
        """
        Put back a connection to the connection pool. The caller is allowed to
        close the connection (if, for instance, it is in an error state), in
        which case it's not returned to the pool and a subsequent get will
        grow the pool if needed.

        :param LDAPConnection conn: the connection managed by the pool.
        :raises ClosedPool: when the method is called on a closed pool.
        :raises PoolError: when tying to put back an object that's not managed
                by this pool.
        """
        if self._closed:
            raise ClosedPool("The pool is closed.")
        try:
            self._used.remove(conn)
            if not conn.closed:
                self._idles.add(conn)
        except KeyError:
            raise PoolError("The %r is not managed by this pool." % conn) from None

    def close(self) -> None:
        """Close the pool and all of its managed connections."""
        for conn in self._idles:
            try:
                conn.close()
            except Exception as exc:
                logger.warning(
                    f"Exception is raised during closing idle connection: {exc}"
                )
        for conn in self._used:
            try:
                conn.close()
            except Exception as exc:
                logger.warning(
                    f"Exception is raised during closing used connection: {exc}"
                )
        self._closed = True
        self._idles = set()
        self._used = set()

    @contextmanager
    def spawn(self, *args: Any, **kwargs: Any) -> Generator[T, None, None]:
        """
        Context manager method that acquires a connection from the pool
        and returns it on exit. It also opens the pool if it hasn't been
        opened before.

        :params \\*args: the positional arguments passed to
                :meth:`bonsai.pool.ConnectionPool.get`.
        :params \\*\\*kwargs: the keyword arguments passed to
                :meth:`bonsai.pool.ConnectionPool.get`.
        """
        conn = None
        try:
            if self._closed:
                self.open()
            conn = self.get(*args, **kwargs)
            yield conn
        finally:
            if conn:
                self.put(conn)

    @property
    def empty(self) -> bool:
        """
        Read-only property that will be True when the connection pool has
        no free connection to use.
        """
        return len(self._idles) == 0 and len(self._used) == self._maxconn

    @property
    def closed(self) -> bool:
        """
        Read-only property that will be True when the connection pool has
        been closed.
        """
        return self._closed

    @property
    def shared_connection(self) -> int:
        """The number of shared connections."""
        return len(self._used)

    @property
    def idle_connection(self) -> int:
        """the number of idle connection."""
        return len(self._idles)

    @property
    def max_connection(self) -> int:
        """The maximal number of connections that the pool can have."""
        return self._maxconn

    @max_connection.setter
    def max_connection(self, val: int) -> None:
        """The maximal number of connections that the pool can have."""
        if val < self._minconn:
            raise ValueError("The maxconn must be greater than minconn.")
        self._maxconn = val


class ThreadedConnectionPool(ConnectionPool[LDAPConnection]):
    """
    A connection pool that can be shared between threads. It's inherited from
    :class:`bonsai.pool.ConnectionPool`.

    :param LDAPClient client: the :class:`bonsai.LDAPClient` that's used to create
                connections.
    :param int minconn: the minimum number of connections that's created
                after the pool is opened.
    :param int maxconn: the maximum number of connections in the pool.
    :param bool block: when it's True, the get method will block when no
                connection is available in the pool.
    :param \\*\\*kwargs: additional keyword arguments that are passed to
                the :meth:`bonsai.LDAPClient.connect` method.
    :raises ValueError: when the minconn is negative or the maxconn is less
        than the minconn.
    """

    def __init__(
        self,
        client: "LDAPClient",
        minconn: int = 1,
        maxconn: int = 10,
        block: bool = True,
        **kwargs: Any,
    ) -> None:
        """Init method."""
        super().__init__(client, minconn, maxconn, **kwargs)
        self._block = block
        self._lock = threading.Condition()

    def get(self, timeout: Optional[float] = None) -> LDAPConnection:
        """
        Get a connection from the connection pool.

        :param float timeout: a timeout until waiting for free connection.
        :raises EmptyPool: when the pool is empty.
        :raises ClosedPool: when the method is called on a closed pool.
        :return: an LDAP connection object.
        """
        with self._lock:
            if self._block:
                self._lock.wait_for(lambda: not self.empty or self._closed, timeout)
            conn = super().get()
            self._lock.notify()
            return conn

    def put(self, conn: LDAPConnection) -> None:
        with self._lock:
            super().put(conn)
            self._lock.notify()

    def close(self) -> None:
        with self._lock:
            super().close()
            self._lock.notify_all()

    def open(self) -> None:
        with self._lock:
            super().open()



import struct
import uuid

from enum import IntEnum
from typing import List, Optional, Set, Union

from .sid import SID


class ACEFlag(IntEnum):
    """ACE type-specific control flags."""

    OBJECT_INHERIT = 0x01
    CONTAINER_INHERIT = 0x02
    NO_PROPAGATE_INHERIT = 0x04
    INHERIT_ONLY = 0x08
    INHERITED = 0x10
    SUCCESSFUL_ACCESS = 0x40
    FAILED_ACCESS = 0x80

    @property
    def short_name(self) -> str:
        """The SDDL short name of the flag."""
        short_names = {
            "OBJECT_INHERIT": "OI",
            "CONTAINER_INHERIT": "CI",
            "NO_PROPAGATE_INHERIT": "NP",
            "INHERIT_ONLY": "IO",
            "INHERITED": "ID",
            "SUCCESSFUL_ACCESS": "SA",
            "FAILED_ACCESS": "FA",
        }
        return short_names[self.name]


class ACEType(IntEnum):
    """Type of the ACE."""

    ACCESS_ALLOWED = 0
    ACCESS_DENIED = 1
    SYSTEM_AUDIT = 2
    SYSTEM_ALARM = 3
    ACCESS_ALLOWED_COMPOUND = 4
    ACCESS_ALLOWED_OBJECT = 5
    ACCESS_DENIED_OBJECT = 6
    SYSTEM_AUDIT_OBJECT = 7
    SYSTEM_ALARM_OBJECT = 8
    ACCESS_ALLOWED_CALLBACK = 9
    ACCESS_DENIED_CALLBACK = 10
    ACCESS_ALLOWED_CALLBACK_OBJECT = 11
    ACCESS_DENIED_CALLBACK_OBJECT = 12
    SYSTEM_AUDIT_CALLBACK = 13
    SYSTEM_ALARM_CALLBACK = 14
    SYSTEM_AUDIT_CALLBACK_OBJECT = 15
    SYSTEM_ALARM_CALLBACK_OBJECT = 16
    SYSTEM_MANDATORY_LABEL = 17
    SYSTEM_RESOURCE_ATTRIBUTE = 18
    SYSTEM_SCOPED_POLICY_ID = 19

    @property
    def short_name(self) -> str:
        """The SDDL short name of the type."""
        short_names = {
            "ACCESS_ALLOWED": "A",
            "ACCESS_DENIED": "D",
            "SYSTEM_AUDIT": "AU",
            "SYSTEM_ALARM": "AL",
            "ACCESS_ALLOWED_COMPOUND": "",
            "ACCESS_ALLOWED_OBJECT": "OA",
            "ACCESS_DENIED_OBJECT": "OD",
            "SYSTEM_AUDIT_OBJECT": "OU",
            "SYSTEM_ALARM_OBJECT": "OL",
            "ACCESS_ALLOWED_CALLBACK": "XA",
            "ACCESS_DENIED_CALLBACK": "XD",
            "ACCESS_ALLOWED_CALLBACK_OBJECT": "ZA",
            "ACCESS_DENIED_CALLBACK_OBJECT": "ZD",
            "SYSTEM_AUDIT_CALLBACK": "XU",
            "SYSTEM_ALARM_CALLBACK": "XL",
            "SYSTEM_AUDIT_CALLBACK_OBJECT": "ZU",
            "SYSTEM_ALARM_CALLBACK_OBJECT": "ZL",
            "SYSTEM_MANDATORY_LABEL": "ML",
            "SYSTEM_RESOURCE_ATTRIBUTE": "RA",
            "SYSTEM_SCOPED_POLICY_ID": "SP",
        }
        return short_names[self.name]

    @property
    def is_object_type(self) -> bool:
        """Flag for ACE types with objects."""
        return self in (
            ACEType.ACCESS_ALLOWED_OBJECT,
            ACEType.ACCESS_DENIED_OBJECT,
            ACEType.SYSTEM_AUDIT_OBJECT,
            ACEType.SYSTEM_ALARM_OBJECT,
            ACEType.ACCESS_ALLOWED_CALLBACK_OBJECT,
            ACEType.ACCESS_DENIED_CALLBACK_OBJECT,
            ACEType.SYSTEM_AUDIT_CALLBACK_OBJECT,
            ACEType.SYSTEM_ALARM_CALLBACK_OBJECT,
        )

    @property
    def is_access_allowed(self) -> bool:
        """Flag for ACE types that allow access."""
        return self in (
            ACEType.ACCESS_ALLOWED,
            ACEType.ACCESS_ALLOWED_COMPOUND,
            ACEType.ACCESS_ALLOWED_OBJECT,
            ACEType.ACCESS_ALLOWED_CALLBACK,
            ACEType.ACCESS_ALLOWED_CALLBACK_OBJECT,
        )

    @property
    def is_access_denied(self) -> bool:
        """Flag for ACE types that deny access."""
        return self in (
            ACEType.ACCESS_DENIED,
            ACEType.ACCESS_DENIED_OBJECT,
            ACEType.ACCESS_DENIED_CALLBACK,
            ACEType.ACCESS_DENIED_CALLBACK_OBJECT,
        )


class ACERight(IntEnum):
    """The rights of the ACE."""

    GENERIC_READ = 0x80000000
    GENERIC_WRITE = 0x4000000
    GENERIC_EXECUTE = 0x20000000
    GENERIC_ALL = 0x10000000
    MAXIMUM_ALLOWED = 0x02000000
    ACCESS_SYSTEM_SECURITY = 0x01000000
    SYNCHRONIZE = 0x00100000
    WRITE_OWNER = 0x00080000
    WRITE_DACL = 0x00040000
    READ_CONTROL = 0x00020000
    DELETE = 0x00010000
    DS_CONTROL_ACCESS = 0x00000100
    DS_CREATE_CHILD = 0x00000001
    DS_DELETE_CHILD = 0x00000002
    ACTRL_DS_LIST = 0x00000004
    DS_SELF = 0x00000008
    DS_READ_PROP = 0x00000010
    DS_WRITE_PROP = 0x00000020
    DS_DELETE_TREE = 0x00000040
    DS_LIST_OBJECT = 0x00000080

    @property
    def short_name(self) -> str:
        """The SDDL short name of the access right."""
        short_names = {
            "GENERIC_READ": "GR",
            "GENERIC_WRITE": "GW",
            "GENERIC_EXECUTE": "GX",
            "GENERIC_ALL": "GA",
            "MAXIMUM_ALLOWED": "MA",
            "ACCESS_SYSTEM_SECURITY": "AS",
            "SYNCHRONIZE": "SY",
            "WRITE_OWNER": "WO",
            "WRITE_DACL": "WD",
            "READ_CONTROL": "RC",
            "DELETE": "SD",
            "DS_CONTROL_ACCESS": "CR",
            "DS_CREATE_CHILD": "CC",
            "DS_DELETE_CHILD": "DC",
            "ACTRL_DS_LIST": "LC",
            "DS_SELF": "SW",
            "DS_READ_PROP": "RP",
            "DS_WRITE_PROP": "WP",
            "DS_DELETE_TREE": "DT",
            "DS_LIST_OBJECT": "LO",
        }
        return short_names[self.name]


class ACLRevision(IntEnum):
    """The ACL revision."""

    ACL_REVISION = 0x02
    ACL_REVISION_DS = 0x04


class ACE:
    """
    A class for the access control entry, that encodes the user rights
    afforded to a principal.

    :param ACEType ace_type: the type of the ACE.
    :param Set[ACEFlag] flags: the set of flags for the ACE.
    :param int mask: the access mask to encode the user rights as an int.
    :param SID trustee_sid: the SID of the trustee.
    :param uuid.UUID|None object_type: a UUID that identifies a property
        set, property, extended right, or type of child object.
    :param uuid.UUID|None inherited_object_type: a UUID that identifies the
        type of child object that can inherit the ACE.
    :param bytes application_data: optional application data.
    """

    def __init__(
        self,
        ace_type: ACEType,
        flags: Set[ACEFlag],
        mask: int,
        trustee_sid: SID,
        object_type: Optional[uuid.UUID],
        inherited_object_type: Optional[uuid.UUID],
        application_data: bytes,
    ) -> None:
        self.__type = ace_type
        self.__flags = flags
        self.__mask = mask
        self.__object_type = object_type
        self.__inherited_object_type = inherited_object_type
        self.__trustee_sid = trustee_sid
        self.__application_data = application_data

    @classmethod
    def from_binary(cls, data: bytes) -> "ACE":
        """
        Create an ACE object from a binary blob.

        :param bytes data: a little-endian byte ordered byte input.
        :returns: A new ACE instance.
        :rtype: ACE
        :raises TypeError: when the parameter is not bytes.
        :raises ValueError: when the input cannot be parsed as an ACE
            object.
        """
        try:
            if not isinstance(data, bytes):
                raise TypeError("The `data` parameter must be bytes")
            object_type = None
            inherited_object_type = None
            application_data = None
            ace_type, flags, size, mask = struct.unpack("<BBHL", data[:8])
            pos = 8
            if ACEType(ace_type).is_object_type:
                obj_flag = struct.unpack("<I", data[8:12])[0]
                pos += 4
                if obj_flag & 0x00000001:
                    object_type = uuid.UUID(bytes_le=data[pos : pos + 16])
                    pos += 16
                if obj_flag & 0x00000002:
                    inherited_object_type = uuid.UUID(bytes_le=data[pos : pos + 16])
                    pos += 16
            trustee_sid = SID(bytes_le=data[pos:])
            pos += trustee_sid.size
            application_data = data[pos:size]
            this = cls(
                ACEType(ace_type),
                {flg for flg in ACEFlag if flags & flg},
                mask,
                trustee_sid,
                object_type,
                inherited_object_type,
                application_data,
            )
            return this
        except struct.error as err:
            raise ValueError(f"Not a valid binary ACE, {err}") from None

    def __eq__(self, other: object) -> bool:
        """
        Check that ACE object equals to the other object. If the `other`
        is also an ACE object, then basically comapre their byte
        representations.

        :param object other: the other object to compare.
        """
        if not isinstance(other, ACE):
            return NotImplemented
        if self.type != other.type or self.flags != other.flags:
            return False
        if self.size != other.size:
            return False
        return self.to_binary() == other.to_binary()

    def __str__(self) -> str:
        """Return the SDDL string representation of the ACE object."""
        flags = "".join(
            flg.short_name for flg in sorted(self.flags, key=lambda f: f.value)
        )
        rights = "".join(
            rgt.short_name for rgt in sorted(self.rights, key=lambda r: r.value)
        )
        object_guid = self.object_type if self.object_type else ""
        inherit_object_guid = (
            self.inherited_object_type if self.inherited_object_type else ""
        )
        sid = (
            self.trustee_sid.sddl_alias
            if self.trustee_sid.sddl_alias
            else str(self.trustee_sid)
        )
        return f"({self.type.short_name};{flags};{rights};{object_guid};{inherit_object_guid};{sid})"

    def to_binary(self) -> bytes:
        """
        Convert ACE object to binary form with little-endian byte order.

        :returns: Bytes of the binary ACE instance
        :rtype: bytes
        """
        size = self.size
        data = bytearray(size)
        struct.pack_into(
            "<BBHL", data, 0, self.type.value, sum(self.flags), size, self.mask
        )
        pos = 8
        if self.type.is_object_type:
            obj_flag = 0x00000001 if self.object_type else 0
            obj_flag |= 0x00000002 if self.inherited_object_type else 0
            struct.pack_into("<L", data, pos, obj_flag)
            pos += 4
            if self.object_type:
                data[pos : pos + 16] = self.object_type.bytes_le
                pos += 16
            if self.inherited_object_type:
                data[pos : pos + 16] = self.inherited_object_type.bytes_le
                pos += 16
        data[pos : pos + self.trustee_sid.size] = self.trustee_sid.bytes_le
        pos += self.trustee_sid.size
        data[pos : pos + size] = self.application_data
        return bytes(data)

    def set_access_rights(self, rights: Union[int, Set[ACERight]]) -> None:
        """
        Set the access rights for the ACE object. The `rights`
        parameter can be a set of ACERight objects or an integer.

        :param int|Set[ACERight] rights: the access rights
        :raises TypeError: if the parameter not an int or not \
        every item of the set is an ACERight object.
        :raises ValueError: if the provided number is not a valid \
        access mask.
        """
        if isinstance(rights, int):
            if rights != sum({rgt for rgt in ACERight if rights & rgt}):
                raise ValueError(
                    f"{rights} is not a valid access rights for an ACE object."
                )
            self.__mask = rights
        elif all(isinstance(rgt, ACERight) for rgt in rights):
            self.__mask = sum(rights)
        else:
            raise TypeError(
                "The rights parameter must be an int or a set of ACERight objects."
            )

    def set_application_data(self, data: bytes) -> None:
        """
        Set application data for the ACE object.

        :param bytes data: the application data.
        :raises TypeError: if the data is not bytes.
        """
        if not isinstance(data, bytes):
            raise TypeError("Application data must be bytes.")
        self.__application_data = data

    def set_flags(self, flags: Set[ACEFlag]) -> None:
        """
        Set flags for the ACE object.

        :param Set[ACEFlag] flags: a set of ACEFlag objects.
        :raises TypeError: if any of the elements in the set \
        is not an ACEFlag object.
        """
        if not all(isinstance(flg, ACEFlag) for flg in flags):
            raise TypeError("All elements in the set must be an ACEFlag.")
        self.__flags = flags

    def set_object_type(self, object_type: Optional[uuid.UUID]) -> None:
        """
        Set the object type for the ACE object.

        :param uuid.UUID|None object_type: the uuid of the object type.
        :raises TypeError: if the parameter is neither an UUID nor None.
        """
        if object_type is not None and not isinstance(object_type, uuid.UUID):
            raise TypeError("Object type must be UUID or None.")
        self.__object_type = object_type

    def set_inherited_object_type(self, object_type: Optional[uuid.UUID]) -> None:
        """
        Set the inherited object type for the ACE object.

        :param uuid.UUID|None object_type: the uuid of the inherited \
        object type.
        :raises TypeError: if the parameter is neither an UUID nor None.
        """
        if object_type is not None and not isinstance(object_type, uuid.UUID):
            raise TypeError("Inherited object type must be UUID or None.")
        self.__inherited_object_type = object_type

    def set_type(self, ace_type: ACEType) -> None:
        """
        Set type for the ACE object.

        :param ACEType ace_type: the type of the ACE object.
        :raises TypeError: if the parameter is not an ACEType.
        """
        if not isinstance(ace_type, ACEType):
            raise TypeError("The type property must be an ACEType.")
        self.__type = ace_type

    def set_trustee_sid(self, sid: SID) -> None:
        """
        Set the trustee SID for the ACE object.

        :param SID sid: the trustee SID of the ACE object.
        :raises TypeError: if the parameter is not a SID.
        """
        if not isinstance(sid, SID):
            raise TypeError("The trustee sid property must be a SID object.")
        self.__trustee_sid = sid

    @property
    def is_inherited(self) -> bool:
        """Return true if the ACE object is inherited."""
        return ACEFlag.INHERITED in self.flags

    @property
    def type(self) -> ACEType:
        """The type of the ACE."""
        return self.__type

    @type.setter
    def type(self, value: ACEType) -> None:
        self.set_type(value)

    @property
    def flags(self) -> Set[ACEFlag]:
        """The flags of the ACE."""
        return self.__flags

    @flags.setter
    def flags(self, value: Set[ACEFlag]) -> None:
        self.set_flags(value)

    @property
    def size(self) -> int:
        """The binary size of ACE in bytes."""
        size = 8
        if self.type.is_object_type:
            size += 4
            if self.object_type:
                size += 16
            if self.inherited_object_type:
                size += 16
        size += self.trustee_sid.size
        size += len(self.application_data)
        return size

    @property
    def mask(self) -> int:
        """The access mask"""
        return self.__mask

    @mask.setter
    def mask(self, value: int) -> None:
        self.set_access_rights(value)

    @property
    def rights(self) -> Set[ACERight]:
        """The set of ACERights based on the access mask."""
        return {rgt for rgt in ACERight if self.mask & rgt}

    @rights.setter
    def rights(self, value: Set[ACERight]) -> None:
        self.set_access_rights(value)

    @property
    def object_type(self) -> Optional[uuid.UUID]:
        """The uuid of the object type."""
        return self.__object_type

    @object_type.setter
    def object_type(self, value: Optional[uuid.UUID]) -> None:
        self.set_object_type(value)

    @property
    def inherited_object_type(self) -> Optional[uuid.UUID]:
        """The uuid of the inherited object type."""
        return self.__inherited_object_type

    @inherited_object_type.setter
    def inherited_object_type(self, value: Optional[uuid.UUID]) -> None:
        self.set_inherited_object_type(value)

    @property
    def trustee_sid(self) -> SID:
        """The sid of the trustee."""
        return self.__trustee_sid

    @trustee_sid.setter
    def trustee_sid(self, value: SID) -> None:
        self.set_trustee_sid(value)

    @property
    def application_data(self) -> bytes:
        """The possible application data."""
        return self.__application_data

    @application_data.setter
    def application_data(self, value: bytes) -> None:
        self.set_application_data(value)


class ACL:
    """
    The access control list (ACL) is used to specify a list of individual
    access control entries (ACEs). An ACL and an array of ACEs comprise a
    complete access control list.

    :param ACLRevision revision: the revision of the ACL.
    :param List[ACE] aces: list of :class:`ACE`.
    """

    def __init__(self, revision: ACLRevision, aces: List[ACE]) -> None:
        self.__revision = revision
        self.__aces = aces

    @classmethod
    def from_binary(cls, data: bytes) -> "ACL":
        """
        Create an ACL object from a binary blob.

        :param bytes data: a little-endian byte ordered byte input.
        :returns: A new ACL instance.
        :rtype: ACL
        :raises TypeError: when the parameter is not bytes.
        :raises ValueError: when the input cannot be parsed as an ACL
            object.
        """
        try:
            if not isinstance(data, bytes):
                raise TypeError("The `data` parameter must be bytes")
            # Unwanted values are the reserved sbz1, size and sbz2.
            rev, _, _, count, _ = struct.unpack("<BBHHH", data[:8])
            pos = 8
            aces = []
            for _ in range(count):
                ace = ACE.from_binary(data[pos:])
                aces.append(ace)
                pos += ace.size
            this = cls(ACLRevision(rev), aces)
            return this
        except struct.error as err:
            raise ValueError(f"Not a valid binary ACL, {err}") from None

    def to_binary(self) -> bytes:
        """
        Convert ACL object to binary form with little-endian byte order.

        :returns: Bytes of the binary ACL instance
        :rtype: bytes
        """
        size = self.size
        data = bytearray(8)
        struct.pack_into("<BBHHH", data, 0, self.revision, 0, size, len(self.aces), 0)
        for ace in self.aces:
            data.extend(ace.to_binary())
        return bytes(data)

    def set_aces(self, aces: List[ACE]) -> None:
        """
        Set a list of ACE object to the ACL.

        :param List[ACE] aces: a list of ACE objects.
        :raises TypeError: if not every element in the list is \
        an ACE object.
        """
        if not all(isinstance(elem, ACE) for elem in aces):
            raise TypeError("All elements in the list must be ACE objects.")
        self.__aces = aces

    @property
    def revision(self) -> ACLRevision:
        """The revision of ACL."""
        return self.__revision

    @property
    def size(self) -> int:
        """The binary size in bytes."""
        return 8 + sum(ace.size for ace in self.aces)

    @property
    def aces(self) -> List[ACE]:
        """The list of :class:`ACE` objects."""
        return self.__aces

    @aces.setter
    def aces(self, value: List[ACE]) -> None:
        self.set_aces(value)



from typing import List, Union

from .ldapclient import LDAPClient
from .ldapurl import LDAPURL


class LDAPReference:
    """
    Object for handling an LDAP reference.

    :param LDAPClient client: a client object.
    :param list references: list of valid LDAP URLs (as string or \
    :class:`LDAPURL` objects).
    """

    def __init__(
        self, client: LDAPClient, references: List[Union[str, LDAPURL]]
    ) -> None:
        if not isinstance(client, LDAPClient):
            raise TypeError("Client parameter must be an LDAPClient.")
        self.__client = client
        self.__refs = []  # type: List[LDAPURL]
        for ref in references:
            if isinstance(ref, str):
                self.__refs.append(LDAPURL(ref))
            elif isinstance(ref, LDAPURL):
                self.__refs.append(ref)
            else:
                raise TypeError("Reference must be string or LDAPURL.")

    @property
    def client(self) -> LDAPClient:
        """The LDAP client."""
        return self.__client

    @client.setter
    def client(self, value: LDAPClient) -> None:
        if not isinstance(value, LDAPClient):
            raise TypeError("Client property must be an LDAPClient.")
        self.__client = value

    @property
    def references(self) -> List[LDAPURL]:
        """The list of LDAPURLs of the references."""
        return self.__refs



import os
import sys
import pathlib
import tempfile

from contextlib import contextmanager

from distutils.errors import CompileError, LinkError

from setuptools.command.build_ext import build_ext
from setuptools import setup, Extension


@contextmanager
def silent_stderr():
    """Shush stderr for receiving unnecessary errors during setup."""
    devnull = open(os.devnull, "w")
    old = os.dup(sys.stderr.fileno())
    os.dup2(devnull.fileno(), sys.stderr.fileno())
    try:
        yield devnull
    finally:
        os.dup2(old, sys.stderr.fileno())


class BuildExt(build_ext):
    """Custom build_ext to test Kerberose capability."""

    def _have_krb5(self, libs: list) -> bool:
        code = """
        #include <krb5.h>
        #include <gssapi/gssapi_krb5.h>

        int main(void) {
            unsigned int ms = 0;
            krb5_context ctx;
            const char *cname = NULL;
            gss_key_value_set_desc store;

            store.count = 0;
            krb5_init_context(&ctx);
            gss_krb5_ccache_name(&ms, cname, NULL);
            return 0;
        }
        """
        with tempfile.TemporaryDirectory() as tmp_dir:
            name = os.path.join(tmp_dir, "test_krb5")
            src_name = name + ".c"
            with open(src_name, "w") as source:
                source.write(code)
            comp = self.compiler
            try:
                with silent_stderr():
                    if "-coverage" in os.getenv("CFLAGS", ""):
                        # If coverage flag is set.
                        libs.append("gcov")
                    comp.link_executable(
                        comp.compile([src_name], output_dir=tmp_dir),
                        name,
                        libraries=libs,
                        library_dirs=self.library_dirs,
                    )
            except (CompileError, LinkError):
                return False
            else:
                return True

    def build_extensions(self) -> None:
        if sys.platform != "win32":
            if self._have_krb5(["krb5", "gssapi"]):
                self.extensions[0].libraries.extend(["krb5", "gssapi"])
                self.extensions[0].define_macros.append(("HAVE_KRB5", 1))
            elif self._have_krb5(["krb5", "gssapi_krb5"]):
                self.extensions[0].libraries.extend(["krb5", "gssapi_krb5"])
                self.extensions[0].define_macros.append(("HAVE_KRB5", 1))
            else:
                print(
                    "INFO: Kerberos headers and libraries are not found."
                    " Additional GSSAPI capabilities won't be installed."
                )
        return super().build_extensions()


SOURCES = [
    "bonsaimodule.c",
    "ldapentry.c",
    "ldapconnectiter.c",
    "ldapconnection.c",
    "ldapmodlist.c",
    "ldap-xplat.c",
    "ldapsearchiter.c",
    "utils.c",
]

DEPENDS = [
    "ldapconnection.h",
    "ldapentry.h",
    "ldapconnectiter.h",
    "ldapmodlist.h",
    "ldapsearchiter.h",
    "ldap-xplat.h",
    "utils.h",
]

MACROS = []
if sys.platform == "darwin":
    MACROS.append(("MACOSX", 1))

if sys.platform == "win32":
    LIBS = ["wldap32", "secur32", "Ws2_32"]
    SOURCES.append("wldap-utf8.c")
    DEPENDS.append("wldap-utf8.h")
    MACROS.append(("WIN32", 1))
else:
    LIBS = ["ldap", "lber"]

SOURCES = [os.path.join("src/_bonsai", x) for x in SOURCES]
DEPENDS = [os.path.join("src/_bonsai", x) for x in DEPENDS]

BONSAI_MODULE = Extension(
    "bonsai._bonsai",
    libraries=LIBS,
    sources=SOURCES,
    depends=DEPENDS,
    define_macros=MACROS,
)

# Get the absolute path to the directory of setup.py.
CURRDIR = pathlib.Path(__file__).resolve().parent

# Get version number from the module's __init__.py file.
with open(CURRDIR / "src" / "bonsai" / "__init__.py") as src:
    VER = [
        line.split('"')[1] for line in src.readlines() if line.startswith("__version__")
    ][0]

setup(
    name="bonsai",
    version=VER,
    description="Python 3 module for accessing LDAP directory servers.",
    author="noirello",
    author_email="noirello@gmail.com",
    ext_modules=[BONSAI_MODULE],
    cmdclass={"build_ext": BuildExt},
    package_dir={"bonsai": "src/bonsai"},
    package_data={"bonsai": ["py.typed"]},
    packages=[
        "bonsai",
        "bonsai.active_directory",
        "bonsai.asyncio",
        "bonsai.gevent",
        "bonsai.tornado",
        "bonsai.trio",
    ],
    include_package_data=True,
)
